@@ -2,22 +2,22 @@
  *      menu-cache.c
  *
  *      Copyright 2008 PCMan <pcman.tw@gmail.com>
+ *      Copyright 2009 Jürgen Hötzel <juergen@archlinux.org>
  *      Copyright 2012-2014 Andriy Grytsenko (LStranger) <andrej@rep.kiev.ua>
  *
- *      This program is free software; you can redistribute it and/or modify
- *      it under the terms of the GNU General Public License as published by
- *      the Free Software Foundation; either version 2 of the License, or
- *      (at your option) any later version.
+ *      This library is free software; you can redistribute it and/or
+ *      modify it under the terms of the GNU Lesser General Public
+ *      License as published by the Free Software Foundation; either
+ *      version 2.1 of the License, or (at your option) any later version.
  *
- *      This program is distributed in the hope that it will be useful,
+ *      This library is distributed in the hope that it will be useful,
  *      but WITHOUT ANY WARRANTY; without even the implied warranty of
- *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
- *      GNU General Public License for more details.
+ *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
+ *      Lesser General Public License for more details.
  *
- *      You should have received a copy of the GNU General Public License
- *      along with this program; if not, write to the Free Software
- *      Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston,
- *      MA 02110-1301, USA.
+ *      You should have received a copy of the GNU Lesser General Public
+ *      License along with this library; if not, write to the Free Software
+ *      Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301  USA
  */
 
 #ifdef HAVE_CONFIG_H
@@ -99,6 +99,7 @@ struct _MenuCacheDir
 {
     MenuCacheItem item;
     GSList* children;
+    guint32 flags;
 };
 
 struct _MenuCacheApp
@@ -108,8 +109,10 @@ struct _MenuCacheApp
     char* exec;
     char* working_dir;
     guint32 show_in_flags;
-/*    char** categories;    */
     guint32 flags;
+    char* try_exec;
+    const char **categories;
+    char* keywords;
 };
 
 struct _MenuCache
@@ -125,6 +128,7 @@ struct _MenuCache
     GThread* thr;
     GCancellable* cancellable;
     guint version;
+    guint reload_id;
     gboolean ready : 1; /* used for sync access */
 };
 
@@ -164,6 +168,18 @@ static void read_dir(GDataInputStream* f, MenuCacheDir* dir, MenuCache* cache,
                      MenuCacheFileDir** all_used_files, int n_all_used_files)
 {
     MenuCacheItem* item;
+    char *line;
+    gsize len;
+
+    /* nodisplay flag */
+    if (cache->version >= 2)
+    {
+        line = g_data_input_stream_read_line(f, &len, cache->cancellable, NULL);
+        if (G_UNLIKELY(line == NULL))
+            return;
+        dir->flags = (guint32)atoi(line);
+        g_free(line);
+    }
 
     /* load child items in the dir */
     while( (item = read_item( f, cache, all_used_files, n_all_used_files )) )
@@ -175,19 +191,61 @@ static void read_dir(GDataInputStream* f, MenuCacheDir* dir, MenuCache* cache,
     }
 
     dir->children = g_slist_reverse( dir->children );
+
+    /* set flag by children if working with old cache generator */
+    if (cache->version == 1)
+    {
+        if (dir->children == NULL)
+            dir->flags = FLAG_IS_NODISPLAY;
+        else if ((line = menu_cache_item_get_file_path(MENU_CACHE_ITEM(dir))) != NULL)
+        {
+            GKeyFile *kf = g_key_file_new();
+            if (g_key_file_load_from_file(kf, line, G_KEY_FILE_NONE, NULL) &&
+                g_key_file_get_boolean(kf, G_KEY_FILE_DESKTOP_GROUP,
+                                       G_KEY_FILE_DESKTOP_KEY_NO_DISPLAY, NULL))
+                dir->flags = FLAG_IS_NODISPLAY;
+            g_key_file_free(kf);
+            g_free(line);
+        }
+    }
+}
+
+static char *_unescape_lf(char *str)
+{
+    char *c, *p = str;
+    gsize len = 0;
+
+    while ((c = strchr(p, '\\')) != NULL)
+    {
+        if (p != &str[len])
+            memmove(&str[len], p, c - p);
+        len += (c - p);
+        if (c[1] == 'n')
+        {
+            str[len++] = '\n';
+            c++;
+        }
+        else if (c != &str[len])
+            str[len++] = *c;
+        p = &c[1];
+    }
+    if (p != &str[len])
+        memmove(&str[len], p, strlen(p) + 1);
+    return str;
 }
 
 static void read_app(GDataInputStream* f, MenuCacheApp* app, MenuCache* cache)
 {
     char *line;
     gsize len;
+    GString *str;
 
     /* generic name */
     line = g_data_input_stream_read_line(f, &len, cache->cancellable, NULL);
     if(G_UNLIKELY(line == NULL))
         return;
     if(G_LIKELY(len > 0))
-        app->generic_name = line;
+        app->generic_name = _unescape_lf(line);
     else
         g_free(line);
 
@@ -196,7 +254,7 @@ static void read_app(GDataInputStream* f, MenuCacheApp* app, MenuCache* cache)
     if(G_UNLIKELY(line == NULL))
         return;
     if(G_LIKELY(len > 0))
-        app->exec = line;
+        app->exec = _unescape_lf(line);
     else
         g_free(line);
 
@@ -213,6 +271,82 @@ static void read_app(GDataInputStream* f, MenuCacheApp* app, MenuCache* cache)
         return;
     app->show_in_flags = (guint32)atol(line);
     g_free(line);
+
+    if (cache->version < 2)
+        return;
+
+    /* TryExec */
+    line = g_data_input_stream_read_line(f, &len, cache->cancellable, NULL);
+    if (G_UNLIKELY(line == NULL))
+        return;
+    if (G_LIKELY(len > 0))
+        app->try_exec = g_strchomp(line);
+    else
+        g_free(line);
+
+    /* Path */
+    line = g_data_input_stream_read_line(f, &len, cache->cancellable, NULL);
+    if (G_UNLIKELY(line == NULL))
+        return;
+    if (G_LIKELY(len > 0))
+        app->working_dir = line;
+    else
+        g_free(line);
+
+    /* Categories */
+    line = g_data_input_stream_read_line(f, &len, cache->cancellable, NULL);
+    if (G_UNLIKELY(line == NULL))
+        return;
+    if (G_LIKELY(len > 0))
+    {
+        const char **x;
+
+        /* split and intern all the strings so categories can be processed
+           later for search doing g_quark_try_string()+g_quark_to_string() */
+        app->categories = x = (const char **)g_strsplit(line, ";", 0);
+        while (*x != NULL)
+        {
+            char *cat = (char *)*x;
+            *x = g_intern_string(cat);
+            g_free(cat);
+            x++;
+        }
+    }
+    g_free(line);
+
+    /* Keywords */
+    str = g_string_new(MENU_CACHE_ITEM(app)->name);
+    if (G_LIKELY(app->exec != NULL))
+    {
+        char *sp = strchr(app->exec, ' ');
+        char *bn = strrchr(app->exec, G_DIR_SEPARATOR);
+
+        g_string_append_c(str, ',');
+        if (bn == NULL && sp == NULL)
+            g_string_append(str, app->exec);
+        else if (bn == NULL || (sp != NULL && sp < bn))
+            g_string_append_len(str, app->exec, sp - app->exec);
+        else if (sp == NULL)
+            g_string_append(str, &bn[1]);
+        else
+            g_string_append_len(str, &bn[1], sp - &bn[1]);
+    }
+    if (app->generic_name != NULL)
+    {
+        g_string_append_c(str, ',');
+        g_string_append(str, app->generic_name);
+    }
+    line = g_data_input_stream_read_line(f, &len, cache->cancellable, NULL);
+    if (G_UNLIKELY(line == NULL))
+        return;
+    if (len > 0)
+    {
+        g_string_append_c(str, ',');
+        g_string_append(str, line);
+    }
+    app->keywords = g_utf8_casefold(str->str, str->len);
+    g_string_free(str, TRUE);
+    g_free(line);
 }
 
 static MenuCacheItem* read_item(GDataInputStream* f, MenuCache* cache,
@@ -265,7 +399,7 @@ static MenuCacheItem* read_item(GDataInputStream* f, MenuCache* cache,
     if(G_UNLIKELY(line == NULL))
         goto _fail;
     if(G_LIKELY(len > 0))
-        item->name = line;
+        item->name = _unescape_lf(line);
     else
         g_free(line);
 
@@ -274,7 +408,7 @@ static MenuCacheItem* read_item(GDataInputStream* f, MenuCache* cache,
     if(G_UNLIKELY(line == NULL))
         goto _fail;
     if(G_LIKELY(len > 0))
-        item->comment = line;
+        item->comment = _unescape_lf(line);
     else
         g_free(line);
 
@@ -439,6 +573,7 @@ void menu_cache_unref(MenuCache* cache)
     MENU_CACHE_LOCK;
     if( g_atomic_int_dec_and_test(&cache->n_ref) )
     {
+        /* g_assert(cache->reload_id != 0); */
         unregister_menu_from_server( cache );
         /* DEBUG("unregister to server"); */
         g_hash_table_remove( hash, cache->menu_name );
@@ -537,6 +672,14 @@ MenuCacheItem* menu_cache_item_ref(MenuCacheItem* item)
     return item;
 }
 
+static gboolean menu_cache_reload_idle(gpointer cache)
+{
+    /* do reload once */
+    if (!g_source_is_destroyed(g_main_current_source()))
+        menu_cache_reload(cache);
+    return FALSE;
+}
+
 typedef struct _CacheReloadNotifier
 {
     MenuCacheReloadNotify func;
@@ -572,10 +715,13 @@ MenuCacheNotifyId menu_cache_add_reload_notify(MenuCache* cache, MenuCacheReload
     MENU_CACHE_LOCK;
     is_first = (cache->root_dir == NULL && cache->notifiers == NULL);
     cache->notifiers = g_slist_concat( cache->notifiers, l );
-    MENU_CACHE_UNLOCK;
     /* reload existing file first so it will be ready right away */
-    if(is_first)
-        menu_cache_reload(cache);
+    if(is_first && cache->reload_id == 0)
+        cache->reload_id = g_idle_add_full(G_PRIORITY_HIGH_IDLE,
+                                           menu_cache_reload_idle,
+                                           menu_cache_ref(cache),
+                                           (GDestroyNotify)menu_cache_unref);
+    MENU_CACHE_UNLOCK;
     return (MenuCacheNotifyId)l;
 }
 
@@ -634,6 +780,11 @@ gboolean menu_cache_reload( MenuCache* cache )
     int i, n;
     int ver_maj, ver_min;
 
+    MENU_CACHE_LOCK;
+    if (cache->reload_id)
+        g_source_remove(cache->reload_id);
+    cache->reload_id = 0;
+    MENU_CACHE_UNLOCK;
     file = g_file_new_for_path(cache->cache_file);
     if(!file)
         return FALSE;
@@ -661,6 +812,7 @@ gboolean menu_cache_reload( MenuCache* cache )
     else
         goto _fail;
 
+    g_debug("menu cache: got file version 1.%d", ver_min);
     /* the second line is menu name */
     line = g_data_input_stream_read_line(f, &len, cache->cancellable, NULL);
     if(G_UNLIKELY(line == NULL))
@@ -772,6 +924,10 @@ gboolean menu_cache_item_unref(MenuCacheItem* item)
         {
             MenuCacheApp* app = MENU_CACHE_APP(item);
             g_free( app->exec );
+            g_free(app->try_exec);
+            g_free(app->working_dir);
+            g_free(app->categories);
+            g_free(app->keywords);
             g_slice_free( MenuCacheApp, app );
         }
     }
@@ -891,7 +1047,7 @@ const char* menu_cache_item_get_file_basename( MenuCacheItem* item )
  */
 const char* menu_cache_item_get_file_dirname( MenuCacheItem* item )
 {
-    return item->file_dir->dir + 1;
+    return item->file_dir ? item->file_dir->dir + 1 : NULL;
 }
 
 /**
@@ -1074,7 +1230,7 @@ MenuCacheItem *menu_cache_find_child_by_name(MenuCacheDir *dir, const char *name
  */
 gboolean menu_cache_dir_is_visible(MenuCacheDir *dir)
 {
-    return (dir->children != NULL);
+    return ((dir->flags & FLAG_IS_NODISPLAY) == 0);
 }
 
 /**
@@ -1109,12 +1265,21 @@ const char* menu_cache_app_get_working_dir( MenuCacheApp* app )
     return app->working_dir;
 }
 
-/*
-char** menu_cache_app_get_categories( MenuCacheApp* app )
+/**
+ * menu_cache_app_get_categories
+ * @app: a menu cache item
+ *
+ * Retrieves list of categories for @app. Returned data are owned by menu
+ * cache and should be not freed by caller.
+ *
+ * Returns: (transfer none): list of categories or %NULL.
+ *
+ * Since: 1.0.0
+ */
+const char * const * menu_cache_app_get_categories(MenuCacheApp* app)
 {
-    return NULL;
+    return app->categories;
 }
-*/
 
 /**
  * menu_cache_app_get_use_terminal
@@ -1161,6 +1326,17 @@ guint32 menu_cache_app_get_show_flags( MenuCacheApp* app )
     return app->show_in_flags;
 }
 
+static gboolean _can_be_exec(MenuCacheApp *app)
+{
+    char *path;
+
+    if (app->try_exec == NULL)
+        return TRUE;
+    path = g_find_program_in_path(app->try_exec);
+    g_free(path);
+    return (path != NULL);
+}
+
 /**
  * menu_cache_app_get_is_visible
  * @app: a menu cache item
@@ -1177,7 +1353,8 @@ gboolean menu_cache_app_get_is_visible( MenuCacheApp* app, guint32 de_flags )
 {
     if(app->flags & FLAG_IS_NODISPLAY)
         return FALSE;
-    return !app->show_in_flags || (app->show_in_flags & de_flags);
+    return (!app->show_in_flags || (app->show_in_flags & de_flags)) &&
+           _can_be_exec(app);
 }
 
 /*
@@ -1677,13 +1854,14 @@ static gpointer menu_cache_loader_thread(gpointer data)
 
 /**
  * menu_cache_lookup
- * @menu_name: a menu cache root
+ * @menu_name: a menu name
  *
- * Searches for connection to menu-cached for @menu_name. If no such
- * connection exists then creates new one. Caller can be notified when
- * cache is ready by adding callback.
+ * Searches for connection to menu-cached for @menu_name. If there is no
+ * such connection exist then creates new one. Caller can be notified
+ * when cache is (re)loaded by adding callback. Caller should check if
+ * the cache is already loaded trying to retrieve its root.
  *
- * See also: menu_cache_add_reload_notify().
+ * See also: menu_cache_add_reload_notify(), menu_cache_item_dup_parent().
  *
  * Returns: (transfer full): menu cache descriptor.
  *
@@ -1728,7 +1906,7 @@ MenuCache* menu_cache_lookup( const char* menu_name )
 
 /**
  * menu_cache_lookup_sync
- * @menu_name: a menu cache root
+ * @menu_name: a menu name
  *
  * Searches for data from menu-cached for @menu_name. If no connection
  * exists yet then creates new one and retrieves all data.
@@ -1918,5 +2096,119 @@ MenuCacheItem *menu_cache_find_item_by_id(MenuCache *cache, const char *id)
     return item;
 }
 
-/* TODO:
-const char **menu_cache_app_get_categories(MenuCacheApp *app) */
\ No newline at end of file
+static GSList* list_app_in_dir_for_cat(MenuCacheDir *dir, GSList *list, const char *id)
+{
+    const char **cat;
+    GSList *l;
+
+    for (l = dir->children; l; l = l->next)
+    {
+        MenuCacheItem *item = MENU_CACHE_ITEM(l->data);
+        switch (item->type)
+        {
+        case MENU_CACHE_TYPE_DIR:
+            list = list_app_in_dir_for_cat(MENU_CACHE_DIR(item), list, id);
+            break;
+        case MENU_CACHE_TYPE_APP:
+            cat = MENU_CACHE_APP(item)->categories;
+            if (cat) while (*cat)
+                if (*cat++ == id)
+                {
+                    list = g_slist_prepend(list, menu_cache_item_ref(item));
+                    break;
+                }
+            break;
+        case MENU_CACHE_TYPE_NONE:
+        case MENU_CACHE_TYPE_SEP:
+            break;
+        }
+    }
+    return list;
+}
+
+/**
+ * menu_cache_list_all_for_category
+ * @cache: a menu cache descriptor
+ * @category: category to list items
+ *
+ * Retrieves list of applications in menu cache which have @category in
+ * their list of categories. The search is case-sensitive. Returned list
+ * should be freed with g_slist_free_full(list, menu_cache_item_unref)
+ * after usage.
+ *
+ * Returns: (transfer full) (element-type MenuCacheItem): list of items.
+ *
+ * Since: 1.0.0
+ */
+GSList *menu_cache_list_all_for_category(MenuCache* cache, const char *category)
+{
+    GQuark q;
+    GSList *list;
+
+    g_return_val_if_fail(cache != NULL && category != NULL, NULL);
+    q = g_quark_try_string(category);
+    if (q == 0)
+        return NULL;
+    MENU_CACHE_LOCK;
+    if (G_UNLIKELY(cache->root_dir == NULL))
+        list = NULL;
+    else
+        list = list_app_in_dir_for_cat(cache->root_dir, NULL, g_quark_to_string(q));
+    MENU_CACHE_UNLOCK;
+    return list;
+}
+
+static GSList* list_app_in_dir_for_kw(MenuCacheDir *dir, GSList *list, const char *kw)
+{
+    GSList *l;
+
+    for (l = dir->children; l; l = l->next)
+    {
+        MenuCacheItem *item = MENU_CACHE_ITEM(l->data);
+        switch (item->type)
+        {
+        case MENU_CACHE_TYPE_DIR:
+            list = list_app_in_dir_for_kw(MENU_CACHE_DIR(item), list, kw);
+            break;
+        case MENU_CACHE_TYPE_APP:
+            if (strstr(MENU_CACHE_APP(item)->keywords, kw) != NULL)
+                list = g_slist_prepend(list, menu_cache_item_ref(item));
+            break;
+        case MENU_CACHE_TYPE_NONE:
+        case MENU_CACHE_TYPE_SEP:
+            break;
+        }
+    }
+    return list;
+}
+
+/**
+ * menu_cache_list_all_for_keyword
+ * @cache: a menu cache descriptor
+ * @keyword: a keyword to search
+ *
+ * Retrieves list of applications in menu cache which have a @keyword
+ * as either a word or part of word in exec command, name, generic name
+ * or defined keywords. The search is case-insensitive. Returned list
+ * should be freed with g_slist_free_full(list, menu_cache_item_unref)
+ * after usage.
+ *
+ * Returns: (transfer full) (element-type MenuCacheItem): list of items.
+ *
+ * Since: 1.0.0
+ */
+GSList *menu_cache_list_all_for_keyword(MenuCache* cache, const char *keyword)
+{
+    char *casefolded = g_utf8_casefold(keyword, -1);
+    GSList *list;
+
+    g_return_val_if_fail(cache != NULL && keyword != NULL, NULL);
+    MENU_CACHE_LOCK;
+    if (G_UNLIKELY(cache->root_dir == NULL))
+        list = NULL;
+    else
+        list = list_app_in_dir_for_kw(cache->root_dir, NULL, casefolded);
+    MENU_CACHE_UNLOCK;
+    g_free(casefolded);
+    return list;
+}
